﻿Imports System.Runtime.CompilerServices
Imports System.ComponentModel
Imports System.Data.Common
Imports System.IO
Imports System.Text
Imports System.Reflection

Namespace System.Windows.Forms

   <EditorBrowsable(EditorBrowsableState.Never)> _
   Public Module FormX

      <Extension()> _
      Public Function GetAllControls(ctl As Control) As List(Of Control)
         Dim allCtls = New List(Of Control)
         allCtls.Add(ctl)
         For i = 0 To Integer.MaxValue
            If i = allCtls.Count Then Exit For
            allCtls.AddRange(allCtls(i).Controls.Cast(Of Control).ToArray)
         Next
         Return allCtls
      End Function

#Region "Form.Register(Dataset)"

      Private _KnownDatasets As New Dictionary(Of Type, DataSet)

      ''' <summary>
      ''' registers all datasets of this type and all bindingsources of the form, to improve performance and to avoid data-redundance
      ''' </summary>
      <Extension()> _
      Public Function Register(Of T As DataSet)(frm As Form, dts As T) As T
         Dim tpDts = dts.GetType
         Dim tmp As DataSet = Nothing
         If _KnownDatasets.TryGetValue(tpDts, tmp) Then
            dts = DirectCast(tmp, T)
         Else
            _KnownDatasets.Add(tpDts, dts)
         End If
         Register = dts
         AddHandler frm.FormClosed, AddressOf Form_Closed
         Dim allCtls = frm.GetAllControls
         Dim dgvSuspend = New GridSuspend(dts, allCtls)
         Dim oldDatasets As New HashSet(Of DataSet)
         Dim shouldReplace As Func(Of Object, Boolean) = Function(src) _
            src IsNot Nothing AndAlso _
            (src Is tpDts OrElse src.GetType Is tpDts AndAlso src IsNot dts)
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         For Each ctl In allCtls.OfType(Of ContainerControl)()
            For Each fld In ctl.GetType.GetFields(bindFlags)
               If fld.FieldType Is tpDts Then
                  oldDatasets.Add(DirectCast(fld.GetValue(ctl), DataSet))
                  fld.SetValue(ctl, dts)
                  Continue For
               End If
               If Not GetType(IComponent).IsAssignableFrom(fld.FieldType) Then Continue For
               Dim prp = fld.FieldType.GetProperty("DataSource")
               If prp.Null Then Continue For
               Dim itm = fld.GetValue(ctl)
               If itm Is Nothing Then Continue For
               If shouldReplace(prp.GetValue(itm, Nothing)) Then prp.SetValue(itm, dts, Nothing)
            Next
         Next
         dgvSuspend.Resume()
         oldDatasets.Remove(dts)
         For Each dts In oldDatasets : dts.Dispose() : Next
      End Function

      Private Class GridSuspend
         'beim Umstöpseln der BindingSources die DGVs abhängen
         Private _Tuples As New List(Of Tuple(Of DataGridView, BindingSource))

         Public Sub New(dts As DataSet, allCtls As List(Of Control))
            If dts.Tables.Cast(Of DataTable).All(Function(tb) tb.Rows.Count = 0) Then Return
            For Each dgv In allCtls.OfType(Of DataGridView)()
               Dim bs = TryCast(dgv.DataSource, BindingSource)
               If bs.Null Then Continue For
               _Tuples.Add(Tuple.Create(dgv, bs))
               dgv.DataSource = Nothing
            Next
         End Sub
         Public Sub [Resume]()
            For Each tpl In _Tuples
               tpl.Item1.DataSource = tpl.Item2
            Next
         End Sub
      End Class

#End Region 'Form.Register(Dataset)

#Region "Winfom-Close-Bugs"

      Private Sub Form_Closed(s As Object, e As EventArgs)
         With DirectCast(s, Form)
            .Validate() 'nochn Bug: DGV kriegt beim Disposen einen ZeilenIndex-Fehler, wenns sich die ZufügeZeile im Edit-Modus befindet
            .SuspendLayout()
            With .Controls
               'Controls könnten sich theor. auch gegenseitig disposen, über Events. dieses hier disposed ganz sicher immer das letzte Element was noch da ist
               While .Count > 0 : .Item(.Count - 1).Dispose() : End While
            End With
         End With
      End Sub

      'Beim Schließen des MainForms wird Form_Closed der anneren Forms garnet mehr aufgerufen. Daher hier alle alle Controls disposen
      Private Sub MainForm_Closed(s As Object, e As EventArgs)
         For Each frm As Form In Application.OpenForms
            Form_Closed(frm, e)
         Next
      End Sub

#End Region 'Winfom-Close-Bugs

   End Module

End Namespace
